# Copyright 2019 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for pyglove.patching."""

import unittest
from pyglove.core import patching
from pyglove.core import symbolic
from pyglove.core import typing as schema


class ConditionalPatchingTest(unittest.TestCase):
  """Conditional patching test."""

  def testPatchOnKey(self):
    """Test `patching.patch_on_key`."""
    d = symbolic.Dict(a=1, b={'a': 2, 'b': 1})
    patching.patch_on_key(d, 'a', 3)
    self.assertEqual(d, {'a': 3, 'b': {'a': 3, 'b': 1}})

    patching.patch_on_key(d, 'a', value_fn=lambda v: v + 1)
    self.assertEqual(d, {'a': 4, 'b': {'a': 4, 'b': 1}})

    with self.assertRaisesRegex(
        ValueError, 'Either `value` or `value_fn` should be specified'):
      patching.patch_on_key(d, 'a', value=1, value_fn=lambda v: v + 1)

  def testPatchOnPath(self):
    """Test `patching.patch_on_path`."""
    d = symbolic.Dict(a=1, b={'a': 2, 'b': 1})
    patching.patch_on_path(d, '.+b', 3)
    self.assertEqual(d, {'a': 1, 'b': {'a': 2, 'b': 3}})

    patching.patch_on_path(d, '.*a', value_fn=lambda v: v + 1)
    self.assertEqual(d, {'a': 2, 'b': {'a': 3, 'b': 3}})

  def testPatchOnValue(self):
    """Test `patching.patch_on_value`."""
    d = symbolic.Dict(a=1, b={'a': 2, 'b': 1})
    patching.patch_on_value(d, 1, 3)
    self.assertEqual(d, {'a': 3, 'b': {'a': 2, 'b': 3}})

    patching.patch_on_value(d, 2, value_fn=lambda v: v * 2)
    self.assertEqual(d, {'a': 3, 'b': {'a': 4, 'b': 3}})

  def testPatchOnType(self):
    """Test `patching.patch_on_type`."""
    d = symbolic.Dict(a='abc', b={'a': 2, 'b': 'def'})
    patching.patch_on_type(d, str, 'foo')
    self.assertEqual(d, {'a': 'foo', 'b': {'a': 2, 'b': 'foo'}})

    patching.patch_on_type(d, int, value_fn=lambda v: v * 2)
    self.assertEqual(d, {'a': 'foo', 'b': {'a': 4, 'b': 'foo'}})

  def testPatchOnMember(self):
    """Test `patching.patch_on_member`."""

    @symbolic.members([
        ('x', schema.Int()),
        ('y', schema.Int()),
    ])
    class A(symbolic.Object):
      pass

    d = symbolic.Dict(a=A(x=1, y=2), x=1)
    patching.patch_on_member(d, A, 'x', 2)
    self.assertEqual(d, {'a': A(x=2, y=2), 'x': 1})

    patching.patch_on_member(d, A, 'y', value_fn=lambda v: v * 2)
    self.assertEqual(d, {'a': A(x=2, y=4), 'x': 1})


class PatcherTest(unittest.TestCase):
  """Patcher test."""

  def testPatch(self):
    """Test patch."""
    @patching.patcher([
        ('v', schema.Int())
    ])
    def p1(unused_src, k, v):
      return {k: v}

    @patching.patcher()
    def p2(unused_src):  # pylint: disable=unused-variable
      return {'': symbolic.Dict()}

    @patching.patcher()
    def p3(unused_src):
      def validate(x):
        if 'a' not in x:
          raise ValueError('Key `a` does not exist.')
      return {}, validate

    # Patch with a single rule.
    self.assertEqual(
        patching.patch(symbolic.Dict(x=1, y=2), {'x': 0, 'z': 3}),
        symbolic.Dict(x=0, y=2, z=3))

    # Patch a list of rules without replacement.
    self.assertEqual(
        patching.patch(symbolic.Dict(), [
            p1(k='a', v=1),  # pylint: disable=no-value-for-parameter
            'p1?b&2',
            'p1?c&v=3',
            'p1?k=d&v=4',
            'p1?a&2'
        ]),
        {
            'a': 2,
            'b': 2,
            'c': 3,
            'd': 4
        })

    # Patch a list of rules with replacement.
    self.assertEqual(
        patching.patch(symbolic.Dict(), [
            p1(k='a', v=1),  # pylint: disable=no-value-for-parameter
            'p1?b&2',
            'p1?c&v=3',
            'p2',
            'p1?k=d&v=4',
            'p1?a&2'
        ]),
        {
            'a': 2,
            'd': 4
        })

    # Patch with a rebind dict as the patching rule.
    self.assertEqual(
        patching.patch(symbolic.Dict(a=1, b=2), {'a': 0, 'c': 3}),
        symbolic.Dict(a=0, b=2, c=3))

    # Patch with a rebind function as the patching rule.
    self.assertEqual(
        patching.patch(
            symbolic.Dict(a=1),
            lambda k, v, p: (v + 1) if isinstance(v, int) else v),
        symbolic.Dict(a=2))

    self.assertEqual(
        patching.patch(symbolic.Dict(a=1), [
            lambda k, v: v
        ]),
        symbolic.Dict(a=1))

    # Patch with a patcher with validator.
    self.assertEqual(
        patching.patch(symbolic.Dict(), [
            p1(k='a', v=1),  # pylint: disable=no-value-for-parameter
            'p1?b&2',
            'p3'
        ]),
        {
            'a': 1,
            'b': 2,
        })

    with self.assertRaisesRegex(
        ValueError, 'Key `a` does not exist.'):
      _ = patching.patch(symbolic.Dict(), [
          'p1?b&2',
          p3()  # pylint: disable=no-value-for-parameter
      ])

    # Nothing to patch.
    self.assertEqual(
        patching.patch(symbolic.Dict(), []), {})

    # Bad patch input.
    with self.assertRaisesRegex(
        TypeError, 'Patching rule .* should be a dict of path to values'):
      patching.patch(symbolic.Dict(), [1])

  def testPatcher(self):
    """Test patcher."""
    # Test patcher without symbolic arguments.
    def assert_patch_equal(src, patcher_or_uri, expected_value):
      if isinstance(patcher_or_uri, patching.Patcher):
        patcher = patcher_or_uri
      else:
        patcher = patching.from_uri(patcher_or_uri)
      self.assertIsInstance(patcher, patching.Patcher)
      dst = patcher.patch(src)
      self.assertIs(src, dst)
      self.assertEqual(dst, expected_value)

    @patching.patcher()
    def set_value1(unused_src, k, v='0'):
      return {
          k: v
      }
    assert_patch_equal(symbolic.Dict(), set_value1(k='a'), {'a': '0'})  # pylint: disable=not-callable, no-value-for-parameter
    assert_patch_equal(symbolic.Dict(), 'set_value1?a&1', {'a': '1'})
    assert_patch_equal(symbolic.Dict(), 'set_value1?a&v=1', {'a': '1'})
    assert_patch_equal(symbolic.Dict(), 'set_value1?k=a&v=1', {'a': '1'})
    self.assertIn('set_value1', patching.patcher_names())

    with self.assertRaisesRegex(
        TypeError, 'The 1st argument of .* must be a symbolic type'):
      set_value1(k='a').patch(1)  # pylint: disable=not-callable, no-value-for-parameter

    # Test patcher with symbolic arguments.
    @patching.patcher([
        ('k', schema.Str()),
        ('v', schema.Int())
    ])
    def set_value2(unused_src, k, v):
      return {
          k: v
      }
    assert_patch_equal(symbolic.Dict(), set_value2(k='a', v=1), {'a': 1})  # pylint: disable=no-value-for-parameter
    assert_patch_equal(symbolic.Dict(), 'set_value2?a&1', {'a': 1})
    assert_patch_equal(symbolic.Dict(), 'set_value2?a&v=1', {'a': 1})
    assert_patch_equal(symbolic.Dict(), 'set_value2?k=a&v=1', {'a': 1})
    self.assertIn('set_value2', patching.patcher_names())

    # Test patcher with replacing source value:
    @patching.patcher([
        ('v', schema.Int())
    ])
    def set_value3(unused_src, v):
      return {'': v}
    self.assertEqual(
        set_value3(v=1).patch(symbolic.Dict()), 1)   # pylint: disable=not-callable, no-value-for-parameter
    self.assertEqual(
        patching.from_uri('set_value3?1').patch(symbolic.Dict()), 1)
    self.assertIn('set_value3', patching.patcher_names())

    # Test patcher with returning rebind fn.
    @patching.patcher()
    def increment(unused_src):
      return lambda k, v: (v + 1) if isinstance(v, int) else v

    self.assertEqual(
        increment().patch(symbolic.Dict(a=1, b=2)), symbolic.Dict(a=2, b=3))  # pylint: disable=not-callable, no-value-for-parameter

    @patching.patcher()
    def no_change(unused_src):
      return lambda k, v: v

    self.assertEqual(
        no_change().patch(symbolic.Dict(a=1, b=2)), symbolic.Dict(a=1, b=2))  # pylint: disable=not-callable, no-value-for-parameter

    # Test patcher with composition of rules.
    @patching.patcher()
    def compound(unused_src):
      return [
          set_value2(k='a', v=1),   # pylint: disable=no-value-for-parameter
          'set_value2?k=b&v=2',
          {
              'c': 3
          }
      ]
    self.assertEqual(
        compound().patch(symbolic.Dict(x=1)), symbolic.Dict(x=1, a=1, b=2, c=3))  # pylint: disable=no-value-for-parameter

    # Test patcher with validation logic.
    @patching.patcher()
    def set_value_with_validate(unused_src):
      def validate(x):
        if 'a' not in x:
          raise ValueError('Key `a` does not exist.')
      return {'a': 1}, validate

    p = set_value_with_validate()  # pylint: disable=not-callable, no-value-for-parameter
    x = p.patch(symbolic.Dict())
    p.validate(x)
    del x['a']
    with self.assertRaisesRegex(
        ValueError, 'Key `a` does not exist.'):
      p.validate(x)

    # Bad patcher function.
    @patching.patcher()
    def set_value4(unused_src):
      return 1

    with self.assertRaisesRegex(
        TypeError,
        'Patching rule .* should be a dict of path to values'):
      set_value4().patch(symbolic.Dict())  # pylint: disable=not-callable, no-value-for-parameter

    # Bad patcher argument spec.
    with self.assertRaisesRegex(
        TypeError,
        '.* cannot be used for constraining Patcher target.'):
      @patching.patcher([
          ('unused_src', schema.Int()),
      ])   # pylint: disable=unused-variable
      def set_value5(unused_src, x):
        del x

    # Bad patcher argument spec.
    with self.assertRaisesRegex(
        TypeError,
        '.* cannot be used for constraining Patcher argument'):
      @patching.patcher([
          ('x', schema.Enum(1, [1, 2]))
      ])   # pylint: disable=unused-variable
      def set_value6(unused_src, x=1):
        del x

    # Bad patcher without target argument.
    with self.assertRaisesRegex(
        TypeError,
        r'Patcher function should have at least 1 argument as patching target'):
      @patching.patcher()
      def set_value7():  # pylint: disable=unused-variable
        return {'': 1}

    # Bad patcher with wrong return type for validator.
    @patching.patcher()
    def set_value8(unused_src):
      return {'': 1}, 1

    with self.assertRaisesRegex(
        TypeError,
        r'The validator returned from patcher .* is not callable'):
      set_value8().patch(symbolic.Dict())  # pylint: disable=no-value-for-parameter

  def testIsPatcherParameterSpec(self):
    """Test for `patching.is_patcher_parameter_spec`."""
    def assert_argument_spec(value_spec):
      self.assertTrue(patching._is_patcher_parameter_spec(value_spec))

    def assert_not_argument_spec(value_spec):
      self.assertFalse(patching._is_patcher_parameter_spec(value_spec))

    assert_argument_spec(schema.Any())
    assert_argument_spec(schema.Int())
    assert_argument_spec(schema.Float())
    assert_argument_spec(schema.Bool())
    assert_argument_spec(schema.Str())
    assert_argument_spec(schema.Enum('a', ['a', 'b']))
    assert_argument_spec(schema.Enum('a', ['a', 'b']))
    assert_argument_spec(schema.List(schema.Int()))
    assert_argument_spec(schema.List(schema.Enum('a', ['a', 'b'])))

    assert_not_argument_spec(schema.Dict())
    assert_not_argument_spec(schema.Type(Exception))
    assert_not_argument_spec(schema.Object(ValueError))
    assert_not_argument_spec(schema.Enum(1, [0, 1]))
    assert_not_argument_spec(schema.List(schema.List(schema.Int())))
    assert_not_argument_spec(schema.Union([schema.Int(), schema.Float()]))

  def testFromUri(self):
    """Test `patching.from_uri`."""
    @patching.patcher([
        ('x', schema.Int())
    ], name='my_patcher')
    def foo(src, x, y='bar'):
      del src, x, y

    self.assertEqual(patching.from_uri('my_patcher?1'), foo(x=1))    # pylint: disable=not-callable, no-value-for-parameter
    self.assertEqual(patching.from_uri('my_patcher?x=1'), foo(x=1))  # pylint: disable=not-callable, no-value-for-parameter

    with self.assertRaisesRegex(
        KeyError, 'Patcher .* is not registered'):
      patching.from_uri('foo')

    with self.assertRaisesRegex(
        ValueError,
        r'Cannot convert \'a\' to int'
        r'.*Patcher=.*foo\', Argument=\'x\''):
      patching.from_uri('my_patcher?x=a').patch(symbolic.Dict())

    with self.assertRaisesRegex(
        KeyError,
        r'Too many positional arguments are provided.'):
      patching.from_uri('my_patcher?1&2&3').patch(symbolic.Dict())

    with self.assertRaisesRegex(
        TypeError,
        r'foo\(\) (missing 1 required positional argument|'
        'takes at least 2 arguments)'):
      patching.from_uri('my_patcher').patch(symbolic.Dict())

  def testParseUri(self):
    """Test `patching.parse_uri`."""
    def test_parse(uri, expected_name, expected_args, expected_kwargs):
      name, args, kwargs = patching.parse_uri(uri)
      self.assertEqual(name, expected_name)
      self.assertEqual(args, expected_args)
      self.assertEqual(kwargs, expected_kwargs)

    test_parse('foo', 'foo', [], {})
    test_parse('foo?', 'foo', [''], {})
    test_parse('foo?1', 'foo', ['1'], {})
    test_parse('foo?1&a:b&c', 'foo', ['1', 'a:b', 'c'], {})
    test_parse('foo?a=1', 'foo', [], {'a': '1'})
    test_parse('foo?a=1&b=2', 'foo', [], {'a': '1', 'b': '2'})
    test_parse('foo?1:2:3&b=2', 'foo', ['1:2:3'], {'b': '2'})

    with self.assertRaisesRegex(
        ValueError,
        '.* is not a valid Patcher name.'):
      patching.parse_uri('1foo')

    with self.assertRaisesRegex(
        ValueError,
        'Invalid argument specification: a-1=2'):
      patching.parse_uri('foo?a-1=2')

    with self.assertRaisesRegex(
        ValueError,
        'Positional argument should be provided before keyword arguments'):
      patching.parse_uri('foo?a=1&2')

  def testParseArgs(self):
    """Test `patching.parse_args`."""
    signature = schema.Signature(
        schema.CallableType.FUNCTION, 'foo', '__main__',
        [
            schema.Argument('src', schema.Any()),
            schema.Argument('x', schema.Int()),
            schema.Argument('y', schema.List(
                schema.Float(min_value=0.0, max_value=1.0)))
        ])
    args, kwargs = patching.parse_args(signature, ['0'], {'y': '0.1:0.5'})
    self.assertEqual(args, [0])
    self.assertEqual(kwargs, {'y': [0.1, 0.5]})

    with self.assertRaisesRegex(
        KeyError, 'Unexpected argument'):
      patching.parse_args(signature, [], {'z': '0.1'})

  def testParseArg(self):
    """Test `patching.parse_arg`."""
     # Test float.
    def parse_arg(value_spec, arg_str):
      return patching.parse_arg('__main__.patcher_a', 'x', value_spec, arg_str)
    self.assertTrue(parse_arg(schema.Bool(), 'True'))
    self.assertTrue(parse_arg(schema.Bool(), 'true'))
    self.assertTrue(parse_arg(schema.Bool(), 'yes'))
    self.assertTrue(parse_arg(schema.Bool(), '1'))
    self.assertFalse(parse_arg(schema.Bool(), 'False'))
    self.assertFalse(parse_arg(schema.Bool(), 'false'))
    self.assertFalse(parse_arg(schema.Bool(), 'no'))
    self.assertFalse(parse_arg(schema.Bool(), '0'))
    with self.assertRaisesRegex(
        ValueError,
        r'Cannot convert \'a\' to bool. '
        r'\(Patcher=\'__main__.patcher_a\', Argument=\'x\'\)'):
      parse_arg(schema.Bool(), 'a')

    with self.assertRaisesRegex(
        ValueError,
        r'Cannot convert \'a\' to float. '
        r'\(Patcher=\'__main__.patcher_a\', Argument=\'x\'\)'):
      parse_arg(schema.Float(), 'a')

    with self.assertRaisesRegex(
        ValueError,
        'Value -1.0 is out of range .*path=__main__.patcher_a.x'):
      parse_arg(schema.Float(min_value=0.), '-1.0')

    # Test int.
    self.assertEqual(parse_arg(schema.Int(), '1'), 1)
    self.assertIsNone(parse_arg(schema.Int().noneable(), 'None'))
    with self.assertRaisesRegex(
        ValueError,
        r'Cannot convert \'a\' to int. '
        r'\(Patcher=\'__main__.patcher_a\', Argument=\'x\'\)'):
      parse_arg(schema.Int(), 'a')

    with self.assertRaisesRegex(
        ValueError,
        'Value -1 is out of range .*path=__main__.patcher_a.x'):
      parse_arg(schema.Int(min_value=0), '-1')

    # Test float.
    self.assertEqual(parse_arg(schema.Float(), '1'), 1.0)
    self.assertIsNone(parse_arg(schema.Float().noneable(), 'None'))
    with self.assertRaisesRegex(
        ValueError,
        r'Cannot convert \'a\' to float. '
        r'\(Patcher=\'__main__.patcher_a\', Argument=\'x\'\)'):
      parse_arg(schema.Float(), 'a')
    with self.assertRaisesRegex(
        ValueError, 'Value -1.0 is out of range'):
      parse_arg(schema.Float(min_value=0.), '-1.0')

    # Test string.
    self.assertEqual(parse_arg(schema.Str(), '1'), '1')
    self.assertEqual(parse_arg(schema.Str(), '"1"'), '1')
    self.assertIsNone(parse_arg(schema.Str().noneable(), 'none'))
    self.assertEqual(parse_arg(schema.Str().noneable(), '"none"'), 'none')

    with self.assertRaisesRegex(
        ValueError,
        'String \'bar\' does not match regular expression '
        '.*path=__main__.patcher_a.x'):
      parse_arg(schema.Str(regex='foo.*'), 'bar')

    with self.assertRaisesRegex(
        ValueError,
        'Unmatched quote for string value: \'"1\'.*'):
      parse_arg(schema.Str(), '"1')

    with self.assertRaisesRegex(
        ValueError,
        'Unmatched quote for string value: \'1"\'.*'):
      parse_arg(schema.Str(), '1"')

    # Test enum.
    self.assertEqual(parse_arg(schema.Enum('a', ['a', 'b']), 'b'), 'b')
    with self.assertRaisesRegex(
        ValueError,
        r'.* cannot be used for Patcher argument. '
        r'Only Enum of string type can be used.'):
      parse_arg(schema.Enum(1, [1, 2]), '1')

    # Test list.
    self.assertEqual(parse_arg(schema.List(schema.Int()), ''), [])
    self.assertEqual(parse_arg(schema.List(schema.Int()), '1'), [1])
    self.assertEqual(parse_arg(schema.List(schema.Int()), '1:2'),
                     [1, 2])
    self.assertEqual(parse_arg(schema.List(schema.Bool()), 'yes:no'),
                     [True, False])
    with self.assertRaisesRegex(
        ValueError,
        r'Cannot convert \'a\' to int. '
        r'\(Patcher=\'__main__.patcher_a\', Argument=\'x\[0\]\'\)'):
      parse_arg(schema.List(schema.Int()), 'a:b')

    # Test other types.
    with self.assertRaisesRegex(
        ValueError,
        r'Dict\(\) cannot be used for Patcher argument'):
      parse_arg(schema.Dict(), '{}')

  def testAllowRepeatedPatcherRegistration(self):
    """Test `patching.allow_repeated_patcher_registration`."""
    patching.patcher()

    patching.allow_repeated_patcher_registration()


@patching.patcher([
    ('value', schema.Int())
])
def update_a(unused_src, value):
  return {'a': value}


class ObjectFactoryTest(unittest.TestCase):
  """ObjectFactory test."""

  def testBaseValueOnly(self):
    """Test creating trainer from base experiment."""
    # Create factory using base value in object form.
    value1 = patching.object_factory(symbolic.Dict, symbolic.Dict(a=1))()
    self.assertEqual(value1, {'a': 1})

    # Create factory using base value in callable form.
    value2 = patching.object_factory(
        symbolic.Dict, lambda: symbolic.Dict(a=1))()
    self.assertEqual(value2, value1)

    # Create factory using base value in file form.
    file_db = {}
    def save_handler(v, filepath):
      file_db[filepath] = v

    def load_handler(filepath):
      return file_db[filepath]

    old_save_handler = symbolic.set_save_handler(save_handler)
    old_load_handler = symbolic.set_load_handler(load_handler)

    filepath = 'myfile.json'
    value1.save(filepath)
    value3 = patching.object_factory(symbolic.Dict, filepath)()
    self.assertEqual(value3, value1)

    with self.assertRaisesRegex(
        TypeError,
        '.* is neither an instance of .*, nor a factory or a path '
        'of JSON file that produces an instance of .*'):
      patching.object_factory(symbolic.Dict, symbolic.List())()

    symbolic.set_save_handler(old_save_handler)
    symbolic.set_load_handler(old_load_handler)

  def testFactoryWithPatchers(self):
    """Test base value + patchers."""
    value = patching.object_factory(
        symbolic.Dict, symbolic.Dict(a=1),
        ['update_a?value=2'])()
    self.assertEqual(value, {'a': 2})

  def testFactoryWithParamsOverride(self):
    """Test base value + params override."""
    # Using dict for `params_override`.
    value = patching.object_factory(
        symbolic.Dict, symbolic.Dict(a={'x': 1, 'y': 2}),
        params_override={
            'a': {
                'x': 2
            }
        })()
    self.assertEqual(value, {'a': {'x': 2, 'y': 2}})

    # Using flattened dict for `params_override`.
    value = patching.object_factory(
        symbolic.Dict, symbolic.Dict(a={'x': 1, 'y': 2}),
        params_override={
            'a.x': 2
        })()
    self.assertEqual(value, {'a': {'x': 2, 'y': 2}})

    # Using string for `params_override`.
    value = patching.object_factory(
        symbolic.Dict, symbolic.Dict(a={'x': 1, 'y': 2}),
        params_override='{"a": {"x": 2}}')()
    self.assertEqual(value, {'a': {'x': 2, 'y': 2}})

    # Using flattened dict string for `params_override`.
    value = patching.object_factory(
        symbolic.Dict, symbolic.Dict(a={'x': 1, 'y': 2}),
        params_override='{"a.x": 2}')()
    self.assertEqual(value, {'a': {'x': 2, 'y': 2}})

    with self.assertRaisesRegex(
        TypeError,
        'Loaded value .* is not an instance of .*'):
      patching.object_factory(
          symbolic.Dict, symbolic.Dict(a=1),
          params_override='1')()

  def testFactoryWithPatchersAndParamsOverride(self):
    """Test base value + patchers + params override."""
    value = patching.object_factory(
        symbolic.Dict, symbolic.Dict(a=1, b=2),
        ['update_a?value=2'],
        params_override={'b': 3, 'c': 0})()
    self.assertEqual(value, {
        'a': 2,
        'b': 3,
        'c': 0
    })


if __name__ == '__main__':
  unittest.main()



# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for base scalars."""

import unittest
from pyglove.ext.scalars import base as scalars


class BasicScalarTest(unittest.TestCase):
  """Test basic scalars."""

  def testMakeScalar(self):
    sv = scalars.make_scalar(scalars.Constant(1))
    self.assertIsInstance(sv, scalars.Scalar)
    self.assertEqual(sv(0), 1)
    self.assertEqual(sv(10), 1)

    sv = scalars.make_scalar(1)
    self.assertIsInstance(sv, scalars.Scalar)
    self.assertIsInstance(sv(0), int)
    self.assertEqual(sv(0), 1)
    self.assertEqual(sv(10), 1)

    sv = scalars.make_scalar(lambda step: step)
    self.assertIsInstance(sv, scalars.Scalar)
    self.assertEqual(sv(1), 1)
    self.assertEqual(sv(10), 10)

  def testCurrentStep(self):
    sv = scalars.STEP * 2
    self.assertEqual(sv(0), 0)
    self.assertEqual(sv(10), 20)


class UnaryOpTest(unittest.TestCase):
  """Tests for unary scalar operators."""

  def testNegation(self):
    sv = -scalars.STEP
    self.assertEqual(sv(1), -1)
    self.assertEqual(sv(2), -2)

  def testFloor(self):
    sv = scalars.Constant(1.6).floor()
    self.assertEqual(sv(0), 1)

  def testCeil(self):
    sv = scalars.Constant(1.6).ceil()
    self.assertEqual(sv(0), 2)

  def testAbs(self):
    sv = abs(scalars.Constant(-1))
    self.assertEqual(sv(0), 1)


class BinaryOpTest(unittest.TestCase):
  """Tests for binary scalar operators."""

  def testAdd(self):
    sv = scalars.Constant(1) + 2
    self.assertEqual(sv(0), 3)

    sv = 2 + scalars.Constant(1)
    self.assertEqual(sv(0), 3)

  def testSubstract(self):
    sv = scalars.Constant(1) - 2
    self.assertEqual(sv(0), -1)

    sv = 2 - scalars.Constant(1)
    self.assertEqual(sv(0), 1)

  def testMultiply(self):
    sv = scalars.Constant(1) * 2
    self.assertEqual(sv(0), 2)

    sv = 2 * scalars.Constant(1)
    self.assertEqual(sv(0), 2)

  def testDivide(self):
    sv = scalars.Constant(1) / 2
    self.assertEqual(sv(0), 0.5)

    sv = 2 / scalars.Constant(1)
    self.assertEqual(sv(0), 2)

  def testFloorDivide(self):
    sv = scalars.Constant(1) // 2
    self.assertEqual(sv(0), 0)

    sv = 2 // scalars.Constant(1)
    self.assertEqual(sv(0), 2)

  def testMod(self):
    sv = scalars.Constant(2) % 3
    self.assertEqual(sv(0), 2)

    sv = 3 % scalars.Constant(2)
    self.assertEqual(sv(0), 1)

  def testPower(self):
    sv = scalars.Constant(2) ** 3
    self.assertEqual(sv(0), 8)

    sv = 3 ** scalars.Constant(2)
    self.assertEqual(sv(0), 9)


if __name__ == '__main__':
  unittest.main()



# Copyright 2019 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for mutators."""

import random
import re
import time
import unittest

import pyglove.core as pg


from pyglove.ext.evolution import mutators


class CoverageTestCase(unittest.TestCase):
  """Class to test coverage of methods with random outputs.

  Usage:

  class MyCoverageTestCase(CoverageTestCase):
    ...
    def testMyFunc(self):
      self.assertEventually(func=MyFunc, allowed=[0, 1, 2], required=[1, 2])
  """

  def assertEventually(  # pylint: disable=g-unreachable-test-method
      self, func, required, allowed, timeout_secs=300.0):
    """Tests that calls to the given function meet required and allowed values.

    Args:
      func: function to test.
      required: iterable of required values. Must be hashable and non-empty.
      allowed: iterable of allowed values. Must be hashable and non-empty.
      timeout_secs: fails if more than this time is required.
    """
    required = set(required)
    assert required
    seen = set()
    start_time = time.time()
    while timeout_secs is None or time.time() - start_time < timeout_secs:
      if seen == required:
        return  # Success.
      value = func()
      if value not in allowed:
        self.fail(msg=(f'Disallowed value: {value}.'))
      if value in required:
        seen.add(value)
    missing = [v for v in required if v not in seen]
    self.fail(
        msg=f'Timed out. Missing values: {str([str(v) for v in missing])}.')


class UniformTest(CoverageTestCase):
  """Tests the Uniform.

  In this test, we generally have specialized tests for the root node because
  the implementation uses a special case to handle it.
  """

  def testOriginalDNARemainsUnchanged(self):
    random.seed()
    dna_spec = pg.geno.manyof(10, [
        pg.geno.constant(), pg.geno.constant(), pg.geno.constant()
    ], distinct=False, sorted=False)
    mutator = mutators.Uniform()
    dna = pg.DNA([0, 2, 1, 1, 2, 0, 0, 1, 2, 1])
    dna.use_spec(dna_spec)
    expected = pg.DNA([0, 2, 1, 1, 2, 0, 0, 1, 2, 1])
    _ = mutator.mutate(dna)
    self.assertEqual(dna, expected)

  def testRootNodeValueMutability(self):
    """Tests that the value of a root node can be mutated."""
    random.seed()
    dna_spec = pg.geno.oneof([pg.geno.constant(), pg.geno.constant()])
    mutator = mutators.Uniform()
    dna = pg.DNA(1)
    dna.use_spec(dna_spec)
    def mutate():
      mutated_dna = mutator.mutate(dna)
      return mutated_dna.value != dna.value
    self.assertEventually(func=mutate, required=[True], allowed=[True, False])

  def testRootNodeBranchMutability(self):
    """Tests that a mutation at the root affects the children."""
    random.seed()
    dna_spec = pg.geno.oneof([
        pg.geno.oneof([pg.geno.constant(), pg.geno.constant()]),
        pg.geno.oneof([pg.geno.constant(), pg.geno.constant()])
    ])
    mutator = mutators.Uniform()
    dna = pg.DNA([(0, [1])])
    dna.use_spec(dna_spec)
    def mutate():
      mutated_dna = mutator.mutate(dna)
      return (
          # This mutation happened at the root.
          mutated_dna.value != dna.value and
          # This mutation also modified the branch.
          mutated_dna.children[0].value != dna.children[0].value)
    self.assertEventually(func=mutate, required=[True], allowed=[True, False])

  def testRootNodeWithoutValueBranchMutability(self):
    """Tests that a value-less root node is mutable."""
    random.seed()
    dna_spec = pg.geno.manyof(2, [
        pg.geno.constant(), pg.geno.constant()
    ], distinct=False, sorted=False)
    mutator = mutators.Uniform()
    dna = pg.DNA([1, 0])
    dna.use_spec(dna_spec)
    def mutate():
      mutated_dna = mutator.mutate(dna)
      # If both children changed, the mutation happened at the root.
      return (mutated_dna.children[0].value != dna.children[0].value and
              mutated_dna.children[1].value != dna.children[1].value)
    self.assertEventually(func=mutate, required=[True], allowed=[True, False])

  def testNonRootNodeValueMutability(self):
    """Tests that the value of a non-root node can be mutated."""
    random.seed()
    dna_spec = pg.geno.manyof(2, [
        pg.geno.constant(), pg.geno.constant()
    ], distinct=False, sorted=False)
    mutator = mutators.Uniform()
    mutator._immutable_root = True  # Prevent root from being mutated.
    dna = pg.DNA([1, 0])
    dna.use_spec(dna_spec)
    def mutate():
      mutated_dna = mutator.mutate(dna)
      return mutated_dna != dna
    self.assertEventually(func=mutate, required=[True], allowed=[True, False])

  def testNonRootNodeBranchMutability(self):
    """Tests that a mutation at a non-root node affects its children."""
    random.seed()
    dna_spec = pg.geno.manyof(2, [
        pg.geno.oneof([pg.geno.constant(), pg.geno.constant()]),
        pg.geno.oneof([pg.geno.constant(), pg.geno.constant()])
    ], distinct=False, sorted=False,)
    mutator = mutators.Uniform()
    mutator._immutable_root = True  # Prevent root from being mutated.
    dna = pg.DNA([(0, [0]), (0, [0])])
    dna.use_spec(dna_spec)
    def mutate():
      mutated_dna = mutator.mutate(dna)
      mutated_nonroot_index = None
      num_different_nonroot_nodes = 0
      for nonroot_index in range(2):
        if (mutated_dna.children[nonroot_index].value !=
            dna.children[nonroot_index].value):
          num_different_nonroot_nodes += 1
          mutated_nonroot_index = nonroot_index
      assert num_different_nonroot_nodes <= 1  # Because root cannot be mutated.
      if num_different_nonroot_nodes == 0:
        # Not the case we are interested in: no first-level nodes were modified.
        return False
      return (
          # Check whether the nonroot node's child has changed too.
          mutated_dna.children[mutated_nonroot_index].children[0].value !=
          dna.children[mutated_nonroot_index].children[0].value)
    self.assertEventually(func=mutate, required=[True], allowed=[True, False])

  def testRootNodeWithNontrivialConditionValueCoverage(self):
    random.seed()
    dna_spec = pg.geno.oneof([
        pg.geno.oneof([pg.geno.constant(), pg.geno.constant()]),
        pg.geno.constant()
    ])
    mutator = mutators.Uniform()
    dna = pg.DNA([(0, 0)])
    dna.use_spec(dna_spec)
    expected_mutated_dnas = [
        pg.DNA([(0, 0)]),
        pg.DNA([(0, 1)]),
        pg.DNA([1]),
    ]
    def mutate():
      mutated_dna = mutator.mutate(dna)
      return expected_mutated_dnas.index(mutated_dna)
    expected_indexes = range(len(expected_mutated_dnas))
    self.assertEventually(
        func=mutate, required=expected_indexes, allowed=expected_indexes)

  def testNonRootNodeWithNontrivialConditionValueCoverage(self):
    random.seed()
    dna_spec = pg.geno.manyof(2, [
        pg.geno.oneof([pg.geno.constant(), pg.geno.constant()]),
        pg.geno.constant()
    ], distinct=False)
    mutator = mutators.Uniform()
    dna = pg.DNA([(0, 0), (0, 0)])
    dna.use_spec(dna_spec)
    expected_mutated_first_child = [
        pg.DNA([(0, 0)]),
        pg.DNA([(0, 1)]),
        pg.DNA([1]),
    ]
    def mutate():
      mutated_dna = mutator.mutate(dna)
      return expected_mutated_first_child.index(mutated_dna.children[0])
    expected_indexes = range(len(expected_mutated_first_child))
    self.assertEventually(
        func=mutate, required=expected_indexes, allowed=expected_indexes)

  def testFloatValueCoverage(self):
    """Tests that all float values can be reached."""
    random.seed()
    dna_spec = pg.geno.floatv(0.0, 1.0)
    mutator = mutators.Uniform()
    dna = pg.DNA(0.2)
    dna.use_spec(dna_spec)
    def mutate():
      mutated_dna = mutator.mutate(dna)
      return round(mutated_dna.value * 10.0)
    expected = range(11)
    self.assertEventually(func=mutate, required=expected, allowed=expected)

  def testChoicesValueCoverage(self):
    """Tests that all choices can be reached."""
    random.seed()
    dna_spec = pg.geno.space([
        pg.geno.manyof(2, [
            pg.geno.constant(), pg.geno.constant(), pg.geno.constant()
        ], distinct=False, sorted=False)
    ])
    mutator = mutators.Uniform()
    dna = pg.DNA([1, 0])
    dna.use_spec(dna_spec)
    def mutate():
      mutated_dna = mutator.mutate(dna)
      return mutated_dna.children[1].value
    expected = [0, 1, 2]
    self.assertEventually(func=mutate, required=expected, allowed=expected)

  def testSortedChoicesValueCoverage(self):
    random.seed()
    dna_spec = pg.geno.manyof(2, [
        pg.geno.constant(), pg.geno.constant(), pg.geno.constant()
    ], distinct=False, sorted=True)
    mutator = mutators.Uniform()
    dna = pg.DNA([0, 1])
    dna.use_spec(dna_spec)
    def mutate():
      mutated_dna = mutator.mutate(dna)
      return (mutated_dna.children[0].value, mutated_dna.children[1].value)
    expected = [(0, 0), (0, 1), (0, 2), (1, 1), (1, 2), (2, 2)]
    self.assertEventually(func=mutate, required=expected, allowed=expected)

  def testNestedSortedChoicesValueCoverage(self):
    random.seed()
    dna_spec = pg.dna_spec(
        pg.oneof([
            pg.manyof(
                2, [1, 2, 3], choices_sorted=True, choices_distinct=False),
            1
        ]))

    mutator = mutators.Uniform()
    dna = pg.DNA((0, [0, 1]))
    dna.use_spec(dna_spec)
    def mutate():
      return mutator.mutate(dna)
    expected = [
        pg.DNA((0, [0, 0])),
        pg.DNA((0, [0, 1])),
        pg.DNA((0, [0, 2])),
        pg.DNA((0, [1, 1])),
        pg.DNA((0, [1, 2])),
        pg.DNA((0, [2, 2])),
        pg.DNA(1)
    ]
    self.assertEventually(func=mutate, required=expected, allowed=expected)

  def testDistinctChoicesValueCoverage(self):
    random.seed()
    dna_spec = pg.geno.manyof(2, [
        pg.geno.constant(), pg.geno.constant(), pg.geno.constant()
    ], distinct=True, sorted=False)
    mutator = mutators.Uniform()
    dna = pg.DNA([0, 2])
    dna.use_spec(dna_spec)
    def mutate():
      mutated_dna = mutator.mutate(dna)
      return (mutated_dna.children[0].value, mutated_dna.children[1].value)
    expected = [(0, 1), (0, 2), (1, 0), (1, 2), (2, 0), (2, 1)]
    self.assertEventually(func=mutate, required=expected, allowed=expected)

  def testNestedDistinctChoicesValueCoverage(self):
    random.seed()
    dna_spec = pg.dna_spec(
        pg.oneof([
            pg.manyof(
                2, [1, 2, 3], choices_distinct=True),
            1
        ]))

    mutator = mutators.Uniform()
    dna = pg.DNA((0, [0, 1]))
    dna.use_spec(dna_spec)
    def mutate():
      return mutator.mutate(dna)
    expected = [
        pg.DNA((0, [0, 1])),
        pg.DNA((0, [1, 0])),
        pg.DNA((0, [0, 2])),
        pg.DNA((0, [2, 0])),
        pg.DNA((0, [1, 2])),
        pg.DNA((0, [2, 1])),
        pg.DNA(1)
    ]
    self.assertEventually(func=mutate, required=expected, allowed=expected)

  def testSortedDistinctChoicesValueCoverage(self):
    random.seed()
    dna_spec = pg.geno.manyof(2, [
        pg.geno.constant(), pg.geno.constant(), pg.geno.constant()
    ], distinct=True, sorted=True)
    mutator = mutators.Uniform()
    dna = pg.DNA([0, 2])
    dna.use_spec(dna_spec)
    def mutate():
      mutated_dna = mutator.mutate(dna)
      return (mutated_dna.children[0].value, mutated_dna.children[1].value)
    expected = [(0, 1), (0, 2), (1, 2)]
    self.assertEventually(func=mutate, required=expected, allowed=expected)

  def testNestedSortedDistinctChoicesValueCoverage(self):
    random.seed()
    dna_spec = pg.dna_spec(
        pg.oneof([
            pg.manyof(
                2, [1, 2, 3], choices_sorted=True, choices_distinct=True),
            1
        ]))

    mutator = mutators.Uniform()
    dna = pg.DNA((0, [0, 1]))
    dna.use_spec(dna_spec)
    def mutate():
      return mutator.mutate(dna)
    expected = [
        pg.DNA((0, [0, 1])),
        pg.DNA((0, [0, 2])),
        pg.DNA((0, [1, 2])),
        pg.DNA(1)
    ]
    self.assertEventually(func=mutate, required=expected, allowed=expected)

  def testDistinctChoicesSubtree(self):
    """Tests conditional subtrees of a node with distinct constraints."""
    random.seed()
    dna_spec = pg.geno.manyof(2, [
        pg.geno.constant(),
        pg.geno.oneof([pg.geno.constant(), pg.geno.constant()])
    ], distinct=True, sorted=False)
    mutator = mutators.Uniform()
    dna = pg.DNA([0, (1, 0)])
    dna.use_spec(dna_spec)
    expected_mutated_dnas = [
        pg.DNA([0, (1, 0)]),
        pg.DNA([0, (1, 1)]),
        pg.DNA([(1, 0), 0]),
        pg.DNA([(1, 1), 0]),
    ]
    def mutate():
      mutated_dna = mutator.mutate(dna)
      return expected_mutated_dnas.index(mutated_dna)
    expected_indexes = range(len(expected_mutated_dnas))
    self.assertEventually(
        func=mutate, required=expected_indexes, allowed=expected_indexes)

  def testTargeting(self):
    random.seed()
    dna_spec = pg.geno.space([
        pg.geno.oneof([   # No hints, so not targeted.
            pg.geno.constant(), pg.geno.constant()
        ]),
        pg.geno.oneof([   # Targeted by regex.
            pg.geno.constant(), pg.geno.constant()
        ], hints='bcde'),
        pg.geno.oneof([   # Not targeted by regex.
            pg.geno.constant(), pg.geno.constant()
        ], hints='abcd')
    ])

    def where_fn(dna):
      """Targets nodes according to DNASpec `hints` tag."""
      if dna.spec.hints is None:
        return False  # No hints to target.
      return re.match('^bc.*', dna.spec.hints)
    mutator = mutators.Uniform(where=where_fn)

    dna = pg.DNA([0, 0, 1])
    dna.use_spec(dna_spec)
    expected_mutated_dnas = [
        pg.DNA([0, 0, 1]),
        pg.DNA([0, 1, 1]),
    ]
    def mutate():
      mutated_dna = mutator.mutate(dna)
      return expected_mutated_dnas.index(mutated_dna)
    expected_indexes = range(len(expected_mutated_dnas))
    self.assertEventually(
        func=mutate, required=expected_indexes, allowed=expected_indexes)

  def testCustomDecisionPoint(self):
    def random_dna(random_generator, previous_dna):
      del previous_dna
      return pg.DNA(value=str(random_generator.randint(0, 3)))

    dna_spec = pg.geno.oneof([
        pg.geno.constant(),
        pg.geno.custom(random_dna_fn=random_dna),
        pg.geno.constant(),
    ])
    mutator = mutators.Uniform()
    dna = pg.DNA(0, spec=dna_spec)
    expected_mutated_dnas = [
        pg.DNA(0),
        pg.DNA((1, '0')),
        pg.DNA((1, '1')),
        pg.DNA((1, '2')),
        pg.DNA((1, '3')),
        pg.DNA(2),
    ]
    def mutate():
      mutated_dna = mutator.mutate(dna)
      return expected_mutated_dnas.index(mutated_dna)
    expected_indexes = range(len(expected_mutated_dnas))
    self.assertEventually(
        func=mutate, required=expected_indexes, allowed=expected_indexes)


class SwapTest(CoverageTestCase):
  """Tests the Swap."""

  def testOriginalDNARemainsUnchanged(self):
    """Tests that a mutable root node is mutated."""
    dna_spec = pg.geno.manyof(
        3, [pg.geno.constant() for _ in range(10)],
        distinct=False, sorted=False)
    mutator = mutators.Swap(seed=1)
    dna = pg.DNA([1, 8, 9])
    dna.use_spec(dna_spec)
    expected = pg.DNA([1, 8, 9])
    _ = mutator.mutate(dna)
    self.assertEqual(dna, expected)

  def testRandomSeed(self):
    dna_spec = pg.geno.manyof(
        3, [pg.geno.constant() for _ in range(10)],
        distinct=False, sorted=False)
    mutator = mutators.Swap(seed=1)
    self.assertEqual(
        mutator.mutate(pg.DNA([1, 8, 9], spec=dna_spec)),
        pg.DNA([9, 8, 1]))
    self.assertEqual(
        mutator.mutate(pg.DNA([1, 8, 9], spec=dna_spec)),
        pg.DNA([8, 1, 9]))

  def testCoverage(self):
    random.seed()
    dna_spec = pg.geno.manyof(
        3, [pg.geno.constant() for _ in range(10)],
        distinct=False, sorted=False)
    mutator = mutators.Swap()
    dna = pg.DNA([1, 8, 9])
    dna.use_spec(dna_spec)
    def mutate():
      mutated_dna = mutator.mutate(dna)
      return (mutated_dna.children[0].value, mutated_dna.children[1].value,
              mutated_dna.children[2].value)
    expected = [(1, 9, 8), (9, 8, 1), (8, 1, 9)]
    self.assertEventually(func=mutate, required=expected, allowed=expected)

  def testHandlesLackOfCandidates(self):
    dna_spec = pg.geno.oneof([pg.geno.constant() for _ in range(10)])
    mutator = mutators.Swap()
    dna = pg.DNA([5])
    dna.use_spec(dna_spec)
    mutated_dna = mutator.mutate(dna)
    self.assertEqual(mutated_dna, dna)

  def testTargeting(self):
    random.seed()
    dna_spec = pg.geno.space([
        pg.geno.manyof(2, [  # No hints, so not targeted.
            pg.geno.constant(), pg.geno.constant()
        ]),
        pg.geno.manyof(2, [  # Targeted by regex.
            pg.geno.constant(), pg.geno.constant()
        ], hints='bcde'),
        pg.geno.manyof(2, [  # Not targeted by regex.
            pg.geno.constant(), pg.geno.constant()
        ], hints='abcd')
    ])

    def where_fn(dna):
      """Targets nodes according to DNASpec `hints` tag."""
      if dna.spec.hints is None:
        return False  # No hints to target.
      return re.match('^bc.*', dna.spec.hints)
    mutator = mutators.Swap(where=where_fn)

    dna = pg.DNA([[0, 1], [0, 1], [0, 1]])
    dna.use_spec(dna_spec)
    mutated_dna = mutator.mutate(dna)
    self.assertEqual(mutated_dna, pg.DNA([[0, 1], [1, 0], [0, 1]]))


if __name__ == '__main__':
  unittest.main()



# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for pyglove.core.typing.key_specs."""

import unittest
from pyglove.core.typing import key_specs as ks


class ConstStrKeyTest(unittest.TestCase):
  """Tests for `ConstStrKey`."""

  def test_basics(self):
    key = ks.ConstStrKey('a')
    self.assertEqual(key, key)
    self.assertEqual(key, 'a')
    self.assertEqual(key.text, 'a')
    self.assertNotEqual(key, 'b')
    self.assertIn(key, {'a': 1})
    self.assertEqual(str(key), 'a')
    self.assertEqual(repr(key), 'a')
    self.assertTrue(key.match('a'))
    self.assertFalse(key.match('b'))

  def test_extend(self):
    self.assertEqual(ks.ConstStrKey('a').extend(ks.ConstStrKey('a')).text, 'a')
    with self.assertRaisesRegex(KeyError,
                                '.* cannot extend .* for keys are different.'):
      ks.ConstStrKey('a').extend(ks.ConstStrKey('b'))

  def test_bad_cases(self):
    with self.assertRaisesRegex(KeyError, '\'.\' cannot be used in key.'):
      ks.ConstStrKey('a.b')


class StrKeyTest(unittest.TestCase):
  """Tests for `StrKey`."""

  def test_basics(self):
    key = ks.StrKey()
    self.assertEqual(key, key)
    self.assertEqual(key, ks.StrKey())
    self.assertTrue(key.match('a'))
    self.assertTrue(key.match('abc'))
    self.assertFalse(key.match(1))

  def test_match_with_regex(self):
    key = ks.StrKey('a.*')
    self.assertTrue(key.match('a1'))
    self.assertTrue(key.match('a'))
    self.assertFalse(key.match('b'))
    self.assertFalse(key.match({}))

  def test_extend(self):
    self.assertIsNone(ks.StrKey().extend(ks.StrKey()).regex)
    self.assertEqual(
        ks.StrKey('a.*').extend(ks.StrKey('a.*')).regex.pattern, 'a.*')

    with self.assertRaisesRegex(
        KeyError, '.* cannot extend .* for keys are different.'):
      ks.StrKey('a.*').extend(ks.StrKey(regex='.*'))


class ListKeyTest(unittest.TestCase):
  """Tests for `ListKey`."""

  def test_basics(self):
    self.assertEqual(ks.ListKey(), ks.ListKey())
    self.assertEqual(ks.ListKey(), ks.ListKey(min_value=0))
    self.assertEqual(
        ks.ListKey(min_value=0, max_value=10),
        ks.ListKey(max_value=10))
    self.assertEqual(
        ks.ListKey(max_value=10), ks.ListKey(max_value=10))
    self.assertNotEqual(
        ks.ListKey(min_value=10), ks.ListKey(min_value=5))
    self.assertNotEqual(
        ks.ListKey(max_value=10), ks.ListKey(max_value=5))
    self.assertNotEqual(
        ks.ListKey(min_value=5), ks.ListKey(max_value=5))

  def test_match_with_unbounded_key(self):
    key = ks.ListKey()
    self.assertEqual(key, key)
    self.assertEqual(key.min_value, 0)
    self.assertIsNone(key.max_value)

    self.assertTrue(key.match(1))
    self.assertTrue(key.match(10000))
    self.assertFalse(key.match('a'))

  def test_match_with_bounded_key(self):
    key = ks.ListKey(min_value=2, max_value=10)
    self.assertTrue(key.match(2))
    self.assertFalse(key.match(0))
    self.assertFalse(key.match(10000))
    self.assertFalse(key.match('a'))

  def test_extend(self):
    key = ks.ListKey()
    self.assertIsNone(key.extend(ks.ListKey()).max_value)
    self.assertEqual(key.extend(ks.ListKey(max_value=10)).max_value, 10)
    with self.assertRaisesRegex(
        TypeError, '.* cannot extend .*: incompatible type.'):
      key.extend(ks.TupleKey(1))

    key = ks.ListKey(min_value=2, max_value=10)
    self.assertEqual(key.extend(ks.ListKey()).min_value, 2)
    self.assertEqual(key.extend(ks.ListKey()).max_value, 10)

    with self.assertRaisesRegex(
        TypeError, '.* cannot extend .*: incompatible type.'):
      key.extend(ks.StrKey())

    with self.assertRaisesRegex(
        TypeError, '.* cannot extend .*: min_value is smaller.'):
      key.extend(ks.ListKey(min_value=3))

    with self.assertRaisesRegex(
        TypeError, '.* cannot extend .*: max_value is greater.'):
      key.extend(ks.ListKey(max_value=5))


class TupleKeyTest(unittest.TestCase):
  """Tests for `TupleKey`."""

  def test_basics(self):
    key = ks.TupleKey(0)
    self.assertEqual(key, key)
    self.assertEqual(ks.TupleKey(0), ks.TupleKey(0))
    self.assertNotEqual(ks.TupleKey(0), ks.TupleKey(1))

  def test_match_with_bounded_key(self):
    key = ks.TupleKey(0)
    self.assertTrue(key.match(0))
    self.assertFalse(key.match(1))
    self.assertFalse(key.match('a'))

  def test_match_with_unbounded_key(self):
    key = ks.TupleKey()
    self.assertTrue(key.match(0))
    self.assertTrue(key.match(1))
    self.assertFalse(key.match('a'))

  def test_extend(self):
    self.assertEqual(ks.TupleKey().extend(ks.TupleKey(1)).index, 1)
    self.assertEqual(ks.TupleKey(0).extend(ks.TupleKey(0)).index, 0)

    with self.assertRaisesRegex(
        TypeError, '.* cannot extend .*: incompatible type.'):
      ks.TupleKey().extend(ks.ListKey(10))

    with self.assertRaisesRegex(
        KeyError, '.* cannot extend .*: unmatched index.'):
      ks.TupleKey(2).extend(ks.TupleKey(1))


if __name__ == '__main__':
  unittest.main()



# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for pyglove.core.typing.typed_missing."""

import unittest

from pyglove.core import object_utils
from pyglove.core.typing import typed_missing
from pyglove.core.typing import value_specs


class MissingValueTest(unittest.TestCase):
  """Tests for typed MissingValue class."""

  def test_eq(self):
    self.assertEqual(
        typed_missing.MissingValue(value_specs.Int()),
        object_utils.MISSING_VALUE)

    self.assertEqual(
        object_utils.MISSING_VALUE,
        typed_missing.MissingValue(value_specs.Int()))

    self.assertEqual(
        typed_missing.MissingValue(value_specs.Int()),
        typed_missing.MissingValue(value_specs.Int()))

    self.assertNotEqual(
        typed_missing.MissingValue(value_specs.Int()),
        typed_missing.MissingValue(value_specs.Int(max_value=1)))

    self.assertNotEqual(
        typed_missing.MissingValue(value_specs.Int()),
        typed_missing.MissingValue(value_specs.Str()))

    m = typed_missing.MissingValue(value_specs.Int())
    self.assertEqual(m, m)

  def test_hash(self):
    self.assertEqual(
        hash(typed_missing.MissingValue(value_specs.Int())),
        hash(typed_missing.MissingValue(value_specs.Float())))

    self.assertEqual(
        hash(typed_missing.MissingValue(value_specs.Int())),
        hash(object_utils.MISSING_VALUE))

    self.assertNotEqual(
        hash(typed_missing.MissingValue(value_specs.Int())),
        hash(1))

  def test_format(self):
    """Test MissingValue.format."""
    self.assertEqual(
        typed_missing.MissingValue(value_specs.Int()).format(compact=True),
        'MISSING_VALUE')

    self.assertEqual(
        typed_missing.MissingValue(value_specs.Int()).format(compact=False),
        'MISSING_VALUE(Int())')


if __name__ == '__main__':
  unittest.main()



# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for math scalars."""

import math
import unittest
from pyglove.ext.scalars import base
from pyglove.ext.scalars import maths as scalars


class MathScalarsTest(unittest.TestCase):
  """Math scalars tests."""

  def assertIsClose(self, x, y):
    assert abs(x - y) < 1e-15, (x, y)

  def testSqrt(self):
    sv = scalars.sqrt(base.STEP)
    self.assertEqual(sv(0), 0)
    self.assertEqual(sv(4), 2)

  def testExp(self):
    sv = scalars.exp(base.STEP)
    self.assertEqual(sv(0), 1)
    self.assertEqual(sv(1), math.e)

  def testLog(self):
    sv = scalars.log(base.STEP, 2)
    self.assertEqual(sv(1), 0)
    self.assertEqual(sv(4), 2)

    sv = scalars.log(16, base.STEP)
    self.assertEqual(sv(2), 4)
    self.assertEqual(sv(4), 2)

  def testCos(self):
    sv = scalars.cos(base.STEP * math.pi / 4)
    self.assertIsClose(sv(0), 1)
    self.assertIsClose(sv(1), math.sqrt(2) / 2)
    self.assertIsClose(sv(2), 0)

  def testSin(self):
    sv = scalars.sin(base.STEP * math.pi / 4)
    self.assertIsClose(sv(0), 0)
    self.assertIsClose(sv(1), math.sqrt(2) / 2)
    self.assertIsClose(sv(2), 1)


class HelperFunctionsTest(unittest.TestCase):
  """Test helper functions for popular scalar schedule."""

  def assertIsClose(self, x, y):
    assert abs(x - y) < 1e-15, (x, y)

  def testLinear(self):
    sv = scalars.linear(10, 1, 6)
    self.assertEqual(sv(0), 1)
    self.assertEqual(sv(1), 1.5)
    self.assertEqual(sv(2), 2)
    self.assertEqual(sv(9), 5.5)
    self.assertEqual(sv(10), 6)

    sv = scalars.linear(10, 6, 1)
    self.assertEqual(sv(0), 6)
    self.assertEqual(sv(1), 5.5)
    self.assertEqual(sv(2), 5)
    self.assertEqual(sv(9), 1.5)
    self.assertEqual(sv(10), 1)

  def testCosineDecay(self):
    sv = scalars.cosine_decay(10, end=0.1)
    self.assertIsClose(sv(0), 1.0)
    self.assertIsClose(sv(6), 0.41094235253127376)
    self.assertIsClose(sv(10), 0.1)

  def testExponentialDecay(self):
    sv = scalars.exponential_decay(0.2, decay_interval=2) * 5
    self.assertIsClose(sv(0), 5.0)
    self.assertIsClose(sv(1), 5.0)
    self.assertIsClose(sv(2), 1.0)
    self.assertIsClose(sv(3), 1.0)
    self.assertIsClose(sv(4), 0.2)
    self.assertIsClose(sv(5), 0.2)
    self.assertIsClose(sv(6), 0.04)
    self.assertIsClose(sv(7), 0.04)

    sv = scalars.exponential_decay(0.2, decay_interval=2, staircase=False) * 5
    self.assertIsClose(sv(0), 5.0)
    self.assertIsClose(sv(1), 2.23606797749979)
    self.assertIsClose(sv(2), 1.0)
    self.assertIsClose(sv(3), 0.447213595499958)
    self.assertIsClose(sv(4), 0.2)
    self.assertIsClose(sv(5), 0.0894427190999916)
    self.assertIsClose(sv(6), 0.04)
    self.assertIsClose(sv(7), 0.01788854381999832)

  def testCyclic(self):
    sv = scalars.cyclic(10) * 2
    self.assertIsClose(sv(0), 2.0)
    self.assertIsClose(sv(2), 1.3090169943749475)
    self.assertIsClose(sv(5), 0.0)
    self.assertIsClose(sv(7), 0.6909830056250524)
    self.assertIsClose(sv(10), 2.0)

    sv = scalars.cyclic(10, math.pi) * 2
    self.assertIsClose(sv(0), 0.0)
    self.assertIsClose(sv(2), 0.6909830056250524)
    self.assertIsClose(sv(5), 2.0)
    self.assertIsClose(sv(7), 1.3090169943749477)
    self.assertIsClose(sv(10), 0.0)


if __name__ == '__main__':
  unittest.main()



# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Step-wise early stopping policies."""

import numbers
from typing import Callable, Dict, Iterable, List, Optional, Tuple, Union
import pyglove.core as pg
from pyglove.ext.early_stopping import base


def step_wise_stopping_predicate_spec():
  """Returns the value spec for step-wise predicate."""
  return pg.typing.Callable(
      [pg.typing.Object(pg.tuning.Measurement),  # Measurement of current trial.
       pg.typing.List(                           # Historical measurements.
           pg.typing.Object(pg.tuning.Measurement))],
      returns=pg.typing.Bool())                  # Should stop or not.


@pg.members([
    ('plan', pg.typing.List(
        pg.typing.Tuple([
            pg.typing.Int(min_value=0),           # Gating step.
            step_wise_stopping_predicate_spec(),  # Step-wise predicate.
        ]), min_size=1),
     ('A list of tuple (step, step-wise stopping predicate) as the step-wise '
      'stopping plan.'))
])
class StepWise(base.EarlyStopingPolicyBase):
  """Step-wise early stopping policy."""

  def _on_bound(self):
    super()._on_bound()
    self._gate_history: List[List[pg.tuning.Measurement]] = [
        [] for _ in range(len(self.plan))]
    self._trial_gate_decision: Dict[int, Tuple[int, bool]] = {}

  def should_stop_early(self, trial: pg.tuning.Trial) -> bool:
    """Returns True if a trial should be stopped early."""
    if not trial.measurements:
      return False

    should_stop = False
    gate_index = self._get_gate_index(trial)
    if gate_index >= 0:
      if trial.id in self._trial_gate_decision:
        decision_gate_index, decision = self._trial_gate_decision[trial.id]
        if decision_gate_index == gate_index:
          return decision
      gate_predicate = self.plan[gate_index][1]
      gate_history = self._gate_history[gate_index]
      m = trial.measurements[-1]
      if gate_predicate(m, gate_history):
        should_stop = True
      gate_history.append(m)
      self._trial_gate_decision[trial.id] = (gate_index, should_stop)
    return should_stop

  def _get_gate_index(self, trial: pg.tuning.Trial) -> int:
    """Gets the index of gate for a trial."""
    previous = None if len(trial.measurements) < 2 else trial.measurements[-2]
    current = trial.measurements[-1]
    for i, (step, _) in enumerate(self.plan):
      if (previous is None or previous.step < step) and current.step >= step:
        return i
    return -1

  def recover(self, history: Iterable[pg.tuning.Trial]):
    """Recovers the policy state based on history."""
    for t in history:
      prev_m = None
      next_gate = 0
      for i, m in enumerate(t.measurements):
        if (next_gate != len(self.plan)
            and (prev_m is None or prev_m.step < self.plan[next_gate][0])
            and m.step >= self.plan[next_gate][0]):
          stopping_decision = False
          gate_history = self._gate_history[next_gate]
          if i == len(t.measurements) - 1:
            if t.status == 'COMPLETED':
              stopping_decision = t.infeasible
            elif t.status == 'PENDING':
              # For the last measurement of a pending trial, we use gate
              # predicate to determine the stopping decision.
              gate_predicate = self.plan[next_gate][1]
              stopping_decision = gate_predicate(t, gate_history)
          gate_history.append(m)
          self._trial_gate_decision[t.id] = (next_gate, stopping_decision)
          next_gate += 1
        prev_m = m


@pg.symbolize
def early_stop_by_value(
    step_values: List[Tuple[
        int,          # Gating step.
        float]],      # Value threshold.
    metric: Union[str, Callable[[pg.tuning.Measurement], float]] = 'reward',
    maximize: bool = True):
  """Step-wise early stopping policy based on the value of reward/metric.

  Example::

    policy = early_stop_by_value([
      # Stop at step 1 if trial reward is less than 0.2.
      (1, 0.2),

      # Stop at step 2 if trial reward is less than 0.8.
      (2, 0.8),
    ])()

  Args:
    step_values: A list of tuple (gating step, value threshold).
        gating step - At which step this rule will be triggered.
        value threshold - A float number indicating the threshold value for
          early stopping.
    metric: Based on which metric the value should be compared against.
      Use str for metric name or a callable object that takes a measurement
      object at a given step as input and returns a float value.
    maximize: If True, reward or metric value below the threshold will be
      stopped, otherwise trials with values above the threshold will be stopped.

  Returns:
    A `StepWise` early stopping policy.
  """
  assert isinstance(step_values, list), step_values
  for v in step_values:
    if (not isinstance(v, tuple)
        or len(v) != 2
        or not isinstance(v[0], int)
        or not isinstance(v[1], numbers.Number)):
      raise ValueError(
          f'Invalid definition in `step_values`: {v}. '
          f'Expect a tuple of 2 elements: '
          f'(step: int, threshold: float).')
  def _cmp(x, y) -> bool:
    return x < y if maximize else x > y

  def _value(m: pg.tuning.Measurement) -> float:
    if isinstance(metric, str):
      return m.reward if metric == 'reward' else m.metrics[metric]
    assert callable(metric), metric
    return metric(m)

  def _make_predicate(threshold: float):
    def _predicate(m: pg.tuning.Measurement, unused_history):
      v = _value(m)
      return _cmp(v, threshold)
    return _predicate

  return StepWise([
      (step, _make_predicate(threshold))
      for step, threshold in step_values])


@pg.symbolize
def early_stop_by_rank(
    step_ranks: List[Tuple[
        int,            # Gating step.
        Union[float,    # Rank percentage in (0.0, 1.0).
              int],     # Absolute rank, below which will be stopped.
        int]],          # Min histogram size at the step to trigger stopping.
    metric: Union[str, Callable[[pg.tuning.Measurement], float]] = 'reward',
    maximize: bool = True) -> StepWise:
  """Step-wise early stopping policy based on the rank of reward/metric.

  Example::

    policy = early_stop_by_rank([
      # Stop at step 1 if accuracy is less than top 80% previous trials at
      # this step, enabled when there are at least 5 previous trials reported
      # at this step.
      (1, 0.8, 5),

      # Stop at step 2 if accuracy is less than top 20% previous trials at
      # this step, enabled when there are at least 10 previous trials reported
      # at this step.
      (2, 0.2, 10),

      # Stop at step 3 if accuracy is less than the 3rd best trial at this step,
      # enabled when there are at least 3 previous trials reported at this step.
      (3, 3, 3)
    ], metric='accuracy')()

  Args:
    step_ranks: A list of tuple (gating step, rank threshold, trigger histogram
      size).
        gating step - At which step this rule will be triggered.
        rank threshold - A float number in range (0, 1) indicating the rank
          percentage or an integer (> 0) indicating the absolute rank as the
          threshold for early stopping.
        trigger historgram size - The minimal number of historical trials
          repoted at current step for this rule to trigger.
    metric: Based on which metric the rank will be computed.
      Use str for metric name or a callable object that takes a measurement
      object at a given step as input and returns a float value.
    maximize: If True, reward or metric value below the threshold will be
      stopped, otherwise trials with values above the threshold will be stopped.

  Returns:
    A `StepWise` early stopping policy.
  """
  assert isinstance(step_ranks, list), step_ranks
  for v in step_ranks:
    if (not isinstance(v, tuple)
        or len(v) != 3
        or not isinstance(v[0], int)
        or not isinstance(v[1], (int, float))
        or not isinstance(v[2], int)):
      raise ValueError(
          f'Invalid definition in `step_ranks`: {v}. '
          f'Expect a tuple of 3 elements: '
          f'(step: int, rank: Union[float, int], trigger_size: int).')
    if isinstance(v[1], float) and (v[1] > 1 or v[1] < 0):
      raise ValueError(
          f'Rank must be within range [0.0, 1.0] when it is percentage '
          f'(float). Encountered: {v[1]} in {v}.')

  def _cmp(x, y) -> bool:
    if y is None:
      return False
    return x < y if maximize else x > y

  def _value(m: pg.tuning.Measurement) -> float:
    if isinstance(metric, str):
      return m.reward if metric == 'reward' else m.metrics[metric]
    assert callable(metric), metric
    return metric(m)

  def _value_by_rank(
      h: List[pg.tuning.Measurement],
      threshold: Union[int, float]) -> Optional[float]:
    if isinstance(threshold, float):
      assert 0.0 <= threshold <= 1.0, threshold
      k = int((len(h) - 1) * threshold)
    else:
      assert isinstance(threshold, int)
      k = threshold - 1
      if k < -len(h) or k >= len(h):
        return None
    return sorted([_value(r) for r in h], reverse=maximize)[k]

  def _make_predicate(rank: Union[int, float], trigger_size: int):
    def _predicate(
        m: pg.tuning.Measurement,
        history: List[pg.tuning.Measurement]):
      if not history or len(history) < trigger_size:
        return False
      return _cmp(_value(m), _value_by_rank(history, rank))
    return _predicate
  return StepWise([(step, _make_predicate(rank, trigger_size))
                   for step, rank, trigger_size in step_ranks])



# Copyright 2019 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for neat."""

import random
import time
import unittest

import pyglove.core as pg
from pyglove.ext.evolution import base
from pyglove.ext.evolution import mutators
from pyglove.ext.evolution import neat_lib as neat


def get_trivial_search_space():
  """Trivial search space.

  Each point in the space is a value in [-1, 1].

  Returns:
    A tunable value.
  """
  return pg.float_value(-1.0, 1.0)


def get_hierarchical_search_space():
  """Sets up hierarchical search space to check NEAT can handle complex search spaces."""
  return pg.oneof([
      pg.oneof([0.1, 0.4, pg.oneof([0.3, 4.0, 6.0])]),
      pg.manyof(3, list(range(5)))
  ])


@pg.members([
    ('seed', pg.typing.Int().noneable()),
])
class TrivialMutator(base.Mutator):
  """Mutator for trivial search space.

  Mutations can only change the value by a small amount.
  """

  def _on_bound(self):
    super()._on_bound()
    self._random = random if self.seed is None else random.Random(self.seed)

  def mutate(self, dna, step):
    del step
    value = dna.value + self._random.uniform(-0.01, 0.01)
    if value < -1.0:
      value = -1.0
    if value > 1.0:
      value = 1.0
    return pg.DNA(value)


def trivial_reward(example):
  """Reward for the trivial search space.

  The reward (i.e. fitness) is the value itself. The goal of the search,
  therefore, is to find the value 1.

  Args:
    example: a materialized value.

  Returns:
    The corresponding reward.
  """
  return example


def get_trivial_hash(search_space, algo):
  hashed_value = 0
  for example, feedback in pg.iter(search_space, 30, algo):
    hashed_value ^= int(example * 1000000)
    feedback(example)
  return hashed_value


class NeatTest(unittest.TestCase):

  def testUtilityComputeDiff(self):
    dna_data_pairs = [((0, 0), (0, 1)), ((0, 0), (1, [0, 1, 2])), ((0, 0), 2),
                      ((1, [0, 1, 2]), (1, [0, 2, 1])), ((1, [0, 1, 2]), 2)]
    corresponding_outputs = [(2, 1, 0), (4, 1, 3), (2, 1, 1), (4, 2, 0),
                             (4, 1, 3)]
    for i, dna_data_pair in enumerate(dna_data_pairs):
      left = pg.DNA(dna_data_pair[0])
      right = pg.DNA(dna_data_pair[1])

      self.assertEqual(
          neat._compute_diff(left, right), corresponding_outputs[i])

  def testIntegration(self):
    # Set up search space.
    search_space = get_trivial_search_space()

    # Search algorithm.
    algo = neat.neat(mutator=TrivialMutator(), population_size=10)

    # Search.
    best_reward = None
    iters = 0
    start_time = time.time()

    while True:
      for example, feedback in pg.iter(search_space, 100, algo):
        reward = trivial_reward(example)
        feedback(reward)
        if best_reward is None or reward > best_reward:
          best_reward = reward
        iters += 1
        if reward >= 1.0:
          break
      if reward >= 1.0:
        break
      if time.time() - start_time > 300.0:
        self.fail('Took too long to find a solution.')

  def testHierarchicalSS(self):
    """Checks if the NEAT algorithm can loop over a complex search space."""
    search_space = get_hierarchical_search_space()
    dna_spec = pg.template(search_space).dna_spec()
    algo = neat.neat(mutator=mutators.Uniform(), population_size=10)
    algo.setup(dna_spec)

    for _ in range(100):
      dna = algo.propose()
      reward = random.uniform(-100.0, 100.0)
      algo.feedback(dna, reward)

  def testPermanence(self):
    """Permanence test for the NEAT generator."""
    search_space = get_trivial_search_space()
    algo = neat.neat(mutator=TrivialMutator(seed=1), population_size=10, seed=1)

    # If a CL causes the following assert to fail, it means that the CL is
    # causing a difference in the behavior of the evolutionary algorithms. If
    # this is expected (e.g. a change in the random number generator), then
    # simply update the hash to the new value.
    self.assertEqual(get_trivial_hash(search_space, algo), 561661)


if __name__ == '__main__':
  unittest.main()
